//
// "$Id: FileChooser2.cxx 9552 2012-05-27 15:44:30Z manolo $"
//
// More fltk3::FileChooser routines.
//
// Copyright 1999-2011 by Michael Sweet.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//

// fabien: ATTENTION: Only Out Of Source Gen. because cxx/H files are autogenerated by fluid.
/** \defgroup group_comdlg Common Dialogs classes and functions
    @{
*/
/** \class fltk3::FileChooser
  The fltk3::FileChooser widget displays a standard file selection
  dialog that supports various selection modes.
  
  \image html fltk3::FileChooser.jpg 
  \image latex  fltk3::FileChooser.jpg "fltk3::FileChooser" width=12cm
  
  The fltk3::FileChooser class also exports several static values
  that may be used to localize or customize the appearance of all file chooser
  dialogs:
  
  <CENTER><TABLE BORDER="1">
  <TR>
  	<TH>Member</TH>
  	<TH>Default value</TH>
  </TR>
  <TR>
  	<TD>add_favorites_label</TD>
  	<TD>"Add to Favorites"</TD>
  </TR>
  <TR>
  	<TD>all_files_label</TD>
  	<TD>"All Files (*)"</TD>
  </TR>
  <TR>
  	<TD>custom_filter_label</TD>
  	<TD>"Custom Filter"</TD>
  </TR>
  <TR>
  	<TD>existing_file_label</TD>
  	<TD>"Please choose an existing file!"</TD>
  </TR>
  <TR>
  	<TD>favorites_label</TD>
  	<TD>"Favorites"</TD>
  </TR>
  <TR>
  	<TD>filename_label</TD>
  	<TD>"Filename:"</TD>
  </TR>
  <TR>
  	<TD>filesystems_label</TD>
  	<TD>"My Computer" (WIN32)<BR>
  	"File Systems" (all others)</TD>
  </TR>
 <TR>
	<TD>hidden_label</TD>
	<TD>"Show hidden files:"</TD>
 </TR>
 <TR>
  	<TD>manage_favorites_label</TD>
  	<TD>"Manage Favorites"</TD>
  </TR>
  <TR>
  	<TD>new_directory_label</TD>
  	<TD>"New Directory?"</TD>
  </TR>
  <TR>
  	<TD>new_directory_tooltip</TD>
  	<TD>"Create a new directory."</TD>
  </TR>
  <TR>
  	<TD>preview_label</TD>
  	<TD>"Preview"</TD>
  </TR>
  <TR>
  	<TD>save_label</TD>
  	<TD>"Save"</TD>
  </TR>
  <TR>
  	<TD>show_label</TD>
  	<TD>"Show:"</TD>
  </TR>
 <TR>
  	<TD>sort</TD>
  	<TD>fltk3::numericsort</TD>
  </TR>
  </TABLE></CENTER>

  The fltk3::FileChooser::sort member specifies the sort function that is
  used when loading the contents of a directory and can be customized
  at run-time.

  The fltk3::FileChooser class also exports the fltk3::FileChooser::newButton
  and fltk3::FileChooser::previewButton widgets so that application developers
  can control their appearance and use.  For more complex customization,
  consider copying the FLTK file chooser code and changing it accordingly.
*/
/** @} */

/** \fn fltk3::FileChooser::FileChooser(const char *pathname, const char *pattern, int type, const char *title)
  The constructor creates the fltk3::FileChooser dialog shown.
  The pathname argument can be a directory name or a
  complete file name (in which case the corresponding file is highlighted
  in the list and in the filename input field.)
  
  The pattern argument can be a NULL
  string or "*" to list all files, or it can be a
  series of descriptions and filter strings separated by tab
  characters (\\t). The format of filters is either
  "Description text (patterns)" or just "patterns". A file chooser
  that provides filters for HTML and image files might look like:
  
  \code
  "HTML Files (*.html)\tImage Files (*.{bmp,gif,jpg,png})"
  \endcode
  
  The file chooser will automatically add the "All Files (*)"
  pattern to the end of the string you pass if you do not provide
  one. The first filter in the string is the default filter.
  
  See the FLTK documentation on fltk3::filename_match()
  for the kinds of pattern strings that are supported.
  
  The type argument can be one of the following:
  
  \li \c SINGLE - allows the user to select a single, existing file.
  \li \c MULTI - allows the user to select one or more existing files.
  \li \c CREATE - allows the user to select a single, existing file or
         specify a new filename.
  \li \c DIRECTORY - allows the user to select a single, existing directory.
  
  The title argument is used to set the title bar text for the
  fltk3::FileChooser window.
*/

/** \var fltk3::FileChooser::newButton
  The "new directory" button is exported so that application developers
  can control the appearance and use.
*/

/** \var fltk3::FileChooser::previewButton
  The "preview" button is exported so that application developers can
  control the appearance and use.
*/

/** \var fltk3::FileChooser::showHiddenButton
 When checked, hidden files (i.e., filename begins with dot) are displayed.
 
 The "showHiddenButton" button is exported so that application developers can
 control its appearance.
 */

/** \fn fltk3::FileChooser::~Fl_File_Chooser()
  Destroys the widget and frees all memory used by it.*/

/** \fn void fltk3::FileChooser::color(fltk3::Color c)
  Sets the background color of the fltk3::FileBrowser list.*/

/** \fn fltk3::Color fltk3::FileChooser::color()
  Gets the background color of the fltk3::FileBrowser list.*/

/** \fn int fltk3::FileChooser::count()
  Returns the number of selected files.*/

/** \fn void fltk3::FileChooser::directory(const char *pathname)
  Sets the current directory.*/

/** \fn const char *fltk3::FileChooser::directory()
  Gets the current directory.*/

/** \fn void fltk3::FileChooser::filter(const char *pattern)
  Sets or gets the current filename filter patterns. The filter
  patterns use fltk3::filename_match().
  Multiple patterns can be used by separating them with tabs, like
  <tt>"*.jpg\t*.png\t*.gif\t*"</tt>. In addition, you can provide
  human-readable labels with the patterns inside parenthesis, like
  <tt>"JPEG Files (*.jpg)\tPNG Files (*.png)\tGIF Files (*.gif)\tAll Files (*)"
  </tt>.

  Use filter(NULL) to show all files.
*/

/** \fn const char *fltk3::FileChooser::filter()
  See void filter(const char *pattern)*/

/** \fn void fltk3::FileChooser::filter_value(int f)
  Sets the current filename filter selection.*/

/** \fn int fltk3::FileChooser::filter_value()
  Gets the current filename filter selection.*/

/** \fn void fltk3::FileChooser::hide()
  Hides the fltk3::FileChooser window.*/

/** \fn void fltk3::FileChooser::iconsize(uchar s)
  Sets the size of the icons in the fltk3::FileBrowser.  By
  default the icon size is set to 1.5 times the textsize().
*/

/** \fn uchar fltk3::FileChooser::iconsize()
  Gets the size of the icons in the fltk3::FileBrowser.  By
  default the icon size is set to 1.5 times the textsize().
*/

/** \fn void fltk3::FileChooser::label(const char *l)
  Sets the title bar text for the fltk3::FileChooser.*/

/** \fn const char *fltk3::FileChooser::label()
  Gets the title bar text for the fltk3::FileChooser.*/

/** \fn void fltk3::FileChooser::ok_label(const char *l)
  Sets the label for the "ok" button in the fltk3::FileChooser.
*/

/** \fn const char *fltk3::FileChooser::ok_label()
  Gets the label for the "ok" button in the fltk3::FileChooser.
*/

/** \fn int fltk3::FileChooser::preview() const
   Returns the current state of the preview box. */

/** \fn void fltk3::FileChooser::rescan()
  Reloads the current directory in the fltk3::FileBrowser.*/

/** \fn void fltk3::FileChooser::show()
  Shows the fltk3::FileChooser window.*/

/** \fn void fltk3::FileChooser::textcolor(fltk3::Color c)
  Sets the current fltk3::FileBrowser text color.*/

/** \fn fltk3::Color fltk3::FileChooser::textcolor()
  Gets the current fltk3::FileBrowser text color.*/

/** \fn void fltk3::FileChooser::textfont(fltk3::Font f)
  Sets the current fltk3::FileBrowser text font.*/

/** \fn fltk3::Font fltk3::FileChooser::textfont()
  Gets the current fltk3::FileBrowser text font.*/

/** \fn void fltk3::FileChooser::textsize(fltk3::Fontsize s)
  Sets the current fltk3::FileBrowser text size.*/

/** \fn fltk3::Fontsize fltk3::FileChooser::textsize()
  Gets the current fltk3::FileBrowser text size.*/

/** \fn void fltk3::FileChooser::type(int t)
  Sets the current type of fltk3::FileChooser.*/

/** \fn int fltk3::FileChooser::type()
  Gets the current type of fltk3::FileChooser.*/

/** \fn void fltk3::FileChooser::value(const char *pathname)
  Sets the current value of the selected file.
*/

/** \fn const char *fltk3::FileChooser::value(int f)
 Gets the current value of the selected file(s).
 \p f is a \c 1-based index into a list of
 file names. The number of selected files is returned by
 fltk3::FileChooser::count().
 
 This sample code loops through all selected files:
 \code
 // Get list of filenames user selected from a MULTI chooser
 for ( int t=1; t<=chooser->count(); t++ ) {
 const char *filename = chooser->value(t);
 ...
 }
 \endcode
 */

/** \fn int fltk3::FileChooser::visible()
  Returns 1 if the fltk3::FileChooser window is visible.*/

/** \fn fltk3::Widget* fltk3::FileChooser::add_extra(fltk3::Widget*)
  Adds extra widget at the bottom of fltk3::FileChooser window.
  Returns pointer for previous extra widget or NULL if not set previously.
  If argument is NULL only remove previous extra widget.

  \note fltk3::FileChooser does \b not delete extra widget in destructor!
     To prevent memory leakage, don't forget to delete unused extra widgets
*/
  /** \fn int fltk3::FileChooser::shown()
    Returns non-zero if the file chooser main window show() has been called (but not hide()
    see fltk3::Window::shown()
  */

  /** \fn void fltk3::FileChooser::callback(void (*cb)(fltk3::FileChooser *, void *), void *d = 0)
    Sets the file chooser callback cb and associated data d */

  /** \fn void fltk3::FileChooser::user_data(void *d)
    Sets the file chooser user data d */

  /** \fn void * fltk3::FileChooser::user_data() const
    Gets the file chooser user data d */

// *** END OF OUT OF SOURCE DOC ***

// Contents:
//
//   fltk3::FileChooser::count()             - Return the number of selected files.
//   fltk3::FileChooser::directory()         - Set the directory in the file chooser.
//   fltk3::FileChooser::filter()            - Set the filter(s) for the chooser.
//   fltk3::FileChooser::newdir()            - Make a new directory.
//   fltk3::FileChooser::value()             - Return a selected filename.
//   fltk3::FileChooser::rescan()            - Rescan the current directory.
//   fltk3::FileChooser::favoritesButtonCB() - Handle favorites selections.
//   fltk3::FileChooser::fileListCB()        - Handle clicks (and double-clicks)
//                                          in the fltk3::FileBrowser.
//   fltk3::FileChooser::fileNameCB()        - Handle text entry in the FileBrowser.
//   fltk3::FileChooser::showChoiceCB()      - Handle show selections.
//   compare_dirnames()                   - Compare two directory names.
//   quote_pathname()                     - Quote a pathname for a menu.
//   unquote_pathname()                   - Unquote a pathname from a menu.
//
//   fltk3::FileChooser::add_extra()         - add extra widget at the bottom, return pointer 
//                                          to previous extra widget or NULL if none,
//                                          If argument is NULL extra widget removed.
//                                          NOTE! file chooser does't delete extra widget in 
//                                          destructor! To prevent memory leakage don't forget
//					    delete unused extra widgets by yourself.
//

//
// Include necessary headers.
//

#include <fltk3/FileChooser.h>
#include <fltk3/filename.h>
#include <fltk3/ask.h>
#include <fltk3/x.h>
#include <fltk3/SharedImage.h>
#include <fltk3/draw.h>

#include <stdio.h>
#include <stdlib.h>
#include "flstring.h"
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#if defined(WIN32) && ! defined (__CYGWIN__)
#  include <direct.h>
#  include <io.h>
// Visual C++ 2005 incorrectly displays a warning about the use of POSIX APIs
// on Windows, which is supposed to be POSIX compliant...
#  define access _access
#  define mkdir _mkdir
// Apparently Borland C++ defines DIRECTORY in <direct.h>, which
// interfers with the fltk3::FileIcon enumeration of the same name.
#  ifdef DIRECTORY
#    undef DIRECTORY
#  endif // DIRECTORY
#else
#  include <unistd.h>
#  include <pwd.h>
#endif /* WIN32 */


//
// File chooser label strings and sort function...
//

fltk3::Preferences	fltk3::FileChooser::prefs_(fltk3::Preferences::USER, "fltk.org", "filechooser");

const char	*fltk3::FileChooser::add_favorites_label = "Add to Favorites";
const char	*fltk3::FileChooser::all_files_label = "All Files (*)";
const char	*fltk3::FileChooser::custom_filter_label = "Custom Filter";
const char	*fltk3::FileChooser::existing_file_label = "Please choose an existing file!";
const char	*fltk3::FileChooser::favorites_label = "Favorites";
const char	*fltk3::FileChooser::filename_label = "Filename:";
#ifdef WIN32
const char	*fltk3::FileChooser::filesystems_label = "My Computer";
#else
const char	*fltk3::FileChooser::filesystems_label = "File Systems";
#endif // WIN32
const char	*fltk3::FileChooser::manage_favorites_label = "Manage Favorites";
const char	*fltk3::FileChooser::new_directory_label = "New Directory?";
const char	*fltk3::FileChooser::new_directory_tooltip = "Create a new directory.";
const char	*fltk3::FileChooser::preview_label = "Preview";
const char	*fltk3::FileChooser::save_label = "Save";
const char	*fltk3::FileChooser::show_label = "Show:";
const char      *fltk3::FileChooser::hidden_label = "Show hidden files";
fltk3::FileSortF	*fltk3::FileChooser::sort = fltk3::numericsort;


//
// Local functions...
//

static int	compare_dirnames(const char *a, const char *b);
static void	quote_pathname(char *, const char *, int);
static void	unquote_pathname(char *, const char *, int);


//
// 'fltk3::FileChooser::count()' - Return the number of selected files.
//

int				// O - Number of selected files
fltk3::FileChooser::count() {
  int		i;		// Looping var
  int		fcount;		// Number of selected files
  const char	*filename;	// Filename in input field or list


  filename = fileName->value();

  if (!(type_ & MULTI)) {
    // Check to see if the file name input field is blank...
    if (!filename || !filename[0]) return 0;
    else return 1;
  }

  for (i = 1, fcount = 0; i <= fileList->size(); i ++)
    if (fileList->selected(i)) {
      // See if this file is a directory...
      // matt: why would we do that? It is perfectly legal to select multiple
      // directories in a DIR chooser. They are visually selected and value(i)
      // returns all of them as expected
      //filename = (char *)fileList->text(i);

      //if (filename[strlen(filename) - 1] != '/')
	fcount ++;
    }

  if (fcount) return fcount;
  else if (!filename || !filename[0]) return 0;
  else return 1;
}


//
// 'fltk3::FileChooser::directory()' - Set the directory in the file chooser.
//

void
fltk3::FileChooser::directory(const char *d)// I - Directory to change to
{
  char	*dirptr;			// Pointer into directory


//  printf("fltk3::FileChooser::directory(\"%s\")\n", d == NULL ? "(null)" : d);

  // NULL == current directory
  if (d == NULL)
    d = ".";

#ifdef WIN32
  // See if the filename contains backslashes...
  char	*slash;				// Pointer to slashes
  char	fixpath[FLTK3_PATH_MAX];			// Path with slashes converted
  if (strchr(d, '\\')) {
    // Convert backslashes to slashes...
    strlcpy(fixpath, d, sizeof(fixpath));

    for (slash = strchr(fixpath, '\\'); slash; slash = strchr(slash + 1, '\\'))
      *slash = '/';

    d = fixpath;
  }
#endif // WIN32

  if (d[0] != '\0')
  {
    // Make the directory absolute...
#if (defined(WIN32) && ! defined(__CYGWIN__))|| defined(__EMX__)
    if (d[0] != '/' && d[0] != '\\' && d[1] != ':')
#else
    if (d[0] != '/' && d[0] != '\\')
#endif /* WIN32 || __EMX__ */
      fltk3::filename_absolute(directory_, d);
    else
      strlcpy(directory_, d, sizeof(directory_));

    // Strip any trailing slash...
    dirptr = directory_ + strlen(directory_) - 1;
    if ((*dirptr == '/' || *dirptr == '\\') && dirptr > directory_)
      *dirptr = '\0';

    // See if we have a trailing .. or . in the filename...
    dirptr = directory_ + strlen(directory_) - 3;
    if (dirptr >= directory_ && strcmp(dirptr, "/..") == 0) {
      // Yes, we have "..", so strip the trailing path...
      *dirptr = '\0';
      while (dirptr > directory_) {
        if (*dirptr == '/') break;
	dirptr --;
      }

      if (dirptr >= directory_ && *dirptr == '/')
        *dirptr = '\0';
    } else if ((dirptr + 1) >= directory_ && strcmp(dirptr + 1, "/.") == 0) {
      // Strip trailing "."...
      dirptr[1] = '\0';
    }
  }
  else
    directory_[0] = '\0';

  if (shown()) {
    // Rescan the directory...
    rescan();
  }
}


//
// 'fltk3::FileChooser::favoritesButtonCB()' - Handle favorites selections.
//

void
fltk3::FileChooser::favoritesButtonCB()
{
  int		v;			// Current selection
  char		pathname[FLTK3_PATH_MAX],		// Pathname
		menuname[FLTK3_PATH_MAX];		// Menu name


  v = favoritesButton->value();

  if (!v) {
    // Add current directory to favorites...
    if (getenv("HOME")) v = favoritesButton->size() - 5;
    else v = favoritesButton->size() - 4;

    sprintf(menuname, "favorite%02d", v);

    prefs_.set(menuname, directory_);
    prefs_.flush();

    quote_pathname(menuname, directory_, sizeof(menuname));
    favoritesButton->add(menuname);

    if (favoritesButton->size() > 104) {
      ((fltk3::MenuItem *)favoritesButton->menu())[0].deactivate();
    }
  } else if (v == 1) {
    // Manage favorites...
    favoritesCB(0);
  } else if (v == 2) {
    // Filesystems/My Computer
    directory("");
  } else {
    unquote_pathname(pathname, favoritesButton->text(v), sizeof(pathname));
    directory(pathname);
  }
}


//
// 'fltk3::FileChooser::favoritesCB()' - Handle favorites dialog.
//

void
fltk3::FileChooser::favoritesCB(fltk3::Widget *w)
					// I - Widget
{
  int		i;			// Looping var
  char		name[32],		// Preference name
		pathname[1024];		// Directory in list


  if (!w) {
    // Load the favorites list...
    favList->clear();
    favList->deselect();

    for (i = 0; i < 100; i ++) {
      // Get favorite directory 0 to 99...
      sprintf(name, "favorite%02d", i);

      prefs_.get(name, pathname, "", sizeof(pathname));

      // Stop on the first empty favorite...
      if (!pathname[0]) break;

      // Add the favorite to the list...
      favList->add(pathname,
                   fltk3::FileIcon::find(pathname, fltk3::FileIcon::DIRECTORY));
    }

    favUpButton->deactivate();
    favDeleteButton->deactivate();
    favDownButton->deactivate();
    favOkButton->deactivate();

    favWindow->hotspot(favList);
    favWindow->show();
  } else if (w == favList) {
    i = favList->value();
    if (i) {
      if (i > 1) favUpButton->activate();
      else favUpButton->deactivate();

      favDeleteButton->activate();

      if (i < favList->size()) favDownButton->activate();
      else favDownButton->deactivate();
    } else {
      favUpButton->deactivate();
      favDeleteButton->deactivate();
      favDownButton->deactivate();
    }
  } else if (w == favUpButton) {
    i = favList->value();

    favList->insert(i - 1, favList->text(i), favList->data(i));
    favList->remove(i + 1);
    favList->select(i - 1);

    if (i == 2) favUpButton->deactivate();

    favDownButton->activate();

    favOkButton->activate();
  } else if (w == favDeleteButton) {
    i = favList->value();

    favList->remove(i);

    if (i > favList->size()) i --;
    favList->select(i);

    if (i < favList->size()) favDownButton->activate();
    else favDownButton->deactivate();

    if (i > 1) favUpButton->activate();
    else favUpButton->deactivate();

    if (!i) favDeleteButton->deactivate();

    favOkButton->activate();
  } else if (w == favDownButton) {
    i = favList->value();

    favList->insert(i + 2, favList->text(i), favList->data(i));
    favList->remove(i);
    favList->select(i + 1);

    if ((i + 1) == favList->size()) favDownButton->deactivate();

    favUpButton->activate();

    favOkButton->activate();
  } else if (w == favOkButton) {
    // Copy the new list over...
    for (i = 0; i < favList->size(); i ++) {
      // Set favorite directory 0 to 99...
      sprintf(name, "favorite%02d", i);

      prefs_.set(name, favList->text(i + 1));
    }

    // Clear old entries as necessary...
    for (; i < 100; i ++) {
      // Clear favorite directory 0 to 99...
      sprintf(name, "favorite%02d", i);

      prefs_.get(name, pathname, "", sizeof(pathname));

      if (pathname[0]) prefs_.set(name, "");
      else break;
    }

    update_favorites();
    prefs_.flush();

    favWindow->hide();
  }
}


//
// 'fltk3::FileChooser::fileListCB()' - Handle clicks (and double-clicks) in the
//                                   fltk3::FileBrowser.
//

void
fltk3::FileChooser::fileListCB()
{
  char	*filename,			// New filename
	pathname[FLTK3_PATH_MAX];			// Full pathname to file


  filename = (char *)fileList->text(fileList->value());
  if (!filename)
    return;

  if (!directory_[0]) {
    strlcpy(pathname, filename, sizeof(pathname));
  } else if (strcmp(directory_, "/") == 0) {
    snprintf(pathname, sizeof(pathname), "/%s", filename);
  } else {
    snprintf(pathname, sizeof(pathname), "%s/%s", directory_, filename);
  }

  if (fltk3::event_clicks()) {
#if (defined(WIN32) && ! defined(__CYGWIN__)) || defined(__EMX__)
    if ((strlen(pathname) == 2 && pathname[1] == ':') ||
        fltk3::_filename_isdir_quick(pathname))
#else
    if (fltk3::_filename_isdir_quick(pathname))
#endif /* WIN32 || __EMX__ */
    {
      // Change directories...
      directory(pathname);

      // Reset the click count so that a click in the same spot won't
      // be treated as a triple-click.  We use a value of -1 because
      // the next click will increment click count to 0, which is what
      // we really want...
      fltk3::event_clicks(-1);
    }
    else
    {
      // Hide the window - picked the file...
      window->hide();
      if (callback_) (*callback_)(this, data_);
    }
  }
  else
  {
    // Check if the user clicks on a directory when picking files;
    // if so, make sure only that item is selected...
    filename = pathname + strlen(pathname) - 1;

    if ((type_ & MULTI) && !(type_ & DIRECTORY)) {
      if (*filename == '/') {
	// Clicked on a directory, deselect everything else...
	int i = fileList->value();
	fileList->deselect();
	fileList->select(i);
      } else {
        // Clicked on a file - see if there are other directories selected...
        int i;
	const char *temp;
	for (i = 1; i <= fileList->size(); i ++) {
	  if (i != fileList->value() && fileList->selected(i)) {
	    temp = fileList->text(i);
	    temp += strlen(temp) - 1;
	    if (*temp == '/') break;	// Yes, selected directory
	  }
	}

        if (i <= fileList->size()) {
	  i = fileList->value();
	  fileList->deselect();
	  fileList->select(i);
	}
      }
    }
    // Strip any trailing slash from the directory name...
    if (*filename == '/') *filename = '\0';

//    puts("Setting fileName from fileListCB...");
    fileName->value(pathname);

    // Update the preview box...
    fltk3::remove_timeout((fltk3::TimeoutHandler)previewCB, this);
    fltk3::add_timeout(1.0, (fltk3::TimeoutHandler)previewCB, this);

    // Do any callback that is registered...
    if (callback_) (*callback_)(this, data_);

    // Activate the OK button as needed...
    if (!fltk3::_filename_isdir_quick(pathname) || (type_ & DIRECTORY))
      okButton->activate();
    else
      okButton->deactivate();
  }
}


//
// 'fltk3::FileChooser::fileNameCB()' - Handle text entry in the FileBrowser.
//

void
fltk3::FileChooser::fileNameCB()
{
  char		*filename,	// New filename
		*slash,		// Pointer to trailing slash
		pathname[FLTK3_PATH_MAX],	// Full pathname to file
		matchname[FLTK3_PATH_MAX];	// Matching filename
  int		i,		// Looping var
		min_match,	// Minimum number of matching chars
		max_match,	// Maximum number of matching chars
		num_files,	// Number of files in directory
		first_line;	// First matching line
  const char	*file;		// File from directory

//  puts("fileNameCB()");
//  printf("Event: %s\n", fl_eventnames[fltk3::event()]);

  // Get the filename from the text field...
  filename = (char *)fileName->value();

  if (!filename || !filename[0]) {
    okButton->deactivate();
    return;
  }

  // Expand ~ and $ variables as needed...
  if (strchr(filename, '~') || strchr(filename, '$')) {
    fltk3::filename_expand(pathname, sizeof(pathname), filename);
    filename = pathname;
    value(pathname);
  }

  // Make sure we have an absolute path...
#if (defined(WIN32) && !defined(__CYGWIN__)) || defined(__EMX__)
  if (directory_[0] != '\0' && filename[0] != '/' &&
      filename[0] != '\\' &&
      !(isalpha(filename[0] & 255) && (!filename[1] || filename[1] == ':'))) {
#else
  if (directory_[0] != '\0' && filename[0] != '/') {
#endif /* WIN32 || __EMX__ */
    fltk3::filename_absolute(pathname, sizeof(pathname), filename);
    value(pathname);
    fileName->mark(fileName->position()); // no selection after expansion
  } else if (filename != pathname) {
    // Finally, make sure that we have a writable copy...
    strlcpy(pathname, filename, sizeof(pathname));
  }

  filename = pathname;

  // Now process things according to the key pressed...
  if (fltk3::event_key() == fltk3::EnterKey || fltk3::event_key() == fltk3::KPEnterKey) {
    // Enter pressed - select or change directory...
#if (defined(WIN32) && ! defined(__CYGWIN__)) || defined(__EMX__)
    if ((isalpha(pathname[0] & 255) && pathname[1] == ':' && !pathname[2]) ||
        (fltk3::_filename_isdir_quick(pathname) &&
	 compare_dirnames(pathname, directory_))) {
#else
    if (fltk3::_filename_isdir_quick(pathname) &&
	compare_dirnames(pathname, directory_)) {
#endif /* WIN32 || __EMX__ */
      directory(pathname);
    } else if ((type_ & CREATE) || access(pathname, 0) == 0) {
      if (!fltk3::_filename_isdir_quick(pathname) || (type_ & DIRECTORY)) {
	// Update the preview box...
	update_preview();

	// Do any callback that is registered...
	if (callback_) (*callback_)(this, data_);

	// Hide the window to signal things are done...
	window->hide();
      }
    } else {
      // File doesn't exist, so beep at and alert the user...
      fltk3::alert("%s",existing_file_label);
    }
  }
  else if (fltk3::event_key() != fltk3::DeleteKey &&
           fltk3::event_key() != fltk3::BackSpaceKey) {
    // Check to see if the user has entered a directory...
    if ((slash = strrchr(pathname, '/')) == NULL)
      slash = strrchr(pathname, '\\');

    if (!slash) return;

    // Yes, change directories if necessary...
    *slash++ = '\0';
    filename = slash;

#if defined(WIN32) || defined(__EMX__)
    if (strcasecmp(pathname, directory_) &&
        (pathname[0] || strcasecmp("/", directory_))) {
#else
    if (strcmp(pathname, directory_) &&
        (pathname[0] || strcasecmp("/", directory_))) {
#endif // WIN32 || __EMX__
      int p = fileName->position();
      int m = fileName->mark();

      directory(pathname);

      if (filename[0]) {
	char tempname[FLTK3_PATH_MAX];

	snprintf(tempname, sizeof(tempname), "%s/%s", directory_, filename);
	fileName->value(tempname);
	strlcpy(pathname, tempname, sizeof(pathname));
      }

      fileName->position(p, m);
    }

    // Other key pressed - do filename completion as possible...
    num_files  = fileList->size();
    min_match  = (int)strlen(filename);
    max_match  = min_match + 1;
    first_line = 0;

    for (i = 1; i <= num_files && max_match > min_match; i ++) {
      file = fileList->text(i);

#if (defined(WIN32) && ! defined(__CYGWIN__)) || defined(__EMX__)
      if (strncasecmp(filename, file, min_match) == 0) {
#else
      if (strncmp(filename, file, min_match) == 0) {
#endif // WIN32 || __EMX__
        // OK, this one matches; check against the previous match
	if (!first_line) {
	  // First match; copy stuff over...
	  strlcpy(matchname, file, sizeof(matchname));
	  max_match = (int)strlen(matchname);

          // Strip trailing /, if any...
	  if (matchname[max_match - 1] == '/') {
	    max_match --;
	    matchname[max_match] = '\0';
	  }

	  // And then make sure that the item is visible
          fileList->topline(i);
	  first_line = i;
	} else {
	  // Succeeding match; compare to find maximum string match...
	  while (max_match > min_match)
#if (defined(WIN32) && ! defined(__CYGWIN__)) || defined(__EMX__)
	    if (strncasecmp(file, matchname, max_match) == 0)
#else
	    if (strncmp(file, matchname, max_match) == 0)
#endif // WIN32 || __EMX__
	      break;
	    else
	      max_match --;

          // Truncate the string as needed...
          matchname[max_match] = '\0';
	}
      }
    }

    // If we have any matches, add them to the input field...
    if (first_line > 0 && min_match == max_match &&
        max_match == (int)strlen(fileList->text(first_line))) {
      // This is the only possible match...
      fileList->deselect(0);
      fileList->select(first_line);
      fileList->redraw();
    } else if (max_match > min_match && first_line) {
      // Add the matching portion...
      fileName->replace(
			(int)(filename - pathname), 
			(int)(filename - pathname + min_match),
                        matchname);

      // Highlight it with the cursor at the end of the selection so
      // s/he can press the right arrow to accept the selection
      // (Tab and End also do this for both cases.)
      fileName->position(
			 (int)(filename - pathname + max_match),
	                 (int)(filename - pathname + min_match));
    } else if (max_match == 0) {
      fileList->deselect(0);
      fileList->redraw();
    }

    // See if we need to enable the OK button...
    if (((type_ & CREATE) || !access(fileName->value(), 0)) &&
        (!fltk3::filename_isdir(fileName->value()) || (type_ & DIRECTORY))) {
      okButton->activate();
    } else {
      okButton->deactivate();
    }
  } else {
    // fltk3::DeleteKey or fltk3::BackSpaceKey
    fileList->deselect(0);
    fileList->redraw();
    if (((type_ & CREATE) || !access(fileName->value(), 0)) &&
        (!fltk3::filename_isdir(fileName->value()) || (type_ & DIRECTORY))) {
      okButton->activate();
    } else {
      okButton->deactivate();
    }
  }
}


//
// 'fltk3::FileChooser::filter()' - Set the filter(s) for the chooser.
//

void
fltk3::FileChooser::filter(const char *p)		// I - Pattern(s)
{
  char		*copyp,				// Copy of pattern
		*start,				// Start of pattern
		*end;				// End of pattern
  int		allfiles;			// Do we have a "*" pattern?
  char		temp[FLTK3_PATH_MAX];			// Temporary pattern string


  // Make sure we have a pattern...
  if (!p || !*p) p = "*";

  // Copy the pattern string...
  copyp = strdup(p);

  // Separate the pattern string as necessary...
  showChoice->clear();

  for (start = copyp, allfiles = 0; start && *start; start = end) {
    end = strchr(start, '\t');
    if (end) *end++ = '\0';

    if (strcmp(start, "*") == 0) {
      showChoice->add(all_files_label);
      allfiles = 1;
    } else {
      quote_pathname(temp, start, sizeof(temp));
      showChoice->add(temp);
      if (strstr(start, "(*)") != NULL) allfiles = 1;
    }
  }

  free(copyp);

  if (!allfiles) showChoice->add(all_files_label);

  showChoice->add(custom_filter_label);
  
  showChoice->value(0);
  showChoiceCB();
}


//
// 'fltk3::FileChooser::newdir()' - Make a new directory.
//

void
fltk3::FileChooser::newdir()
{
  const char	*dir;		// New directory name
  char		pathname[FLTK3_PATH_MAX];	// Full path of directory


  // Get a directory name from the user
  if ((dir = fltk3::input("%s", NULL, new_directory_label)) == NULL)
    return;

  // Make it relative to the current directory as needed...
#if (defined(WIN32) && ! defined (__CYGWIN__)) || defined(__EMX__)
  if (dir[0] != '/' && dir[0] != '\\' && dir[1] != ':')
#else
  if (dir[0] != '/' && dir[0] != '\\')
#endif /* WIN32 || __EMX__ */
    snprintf(pathname, sizeof(pathname), "%s/%s", directory_, dir);
  else
    strlcpy(pathname, dir, sizeof(pathname));

  // Create the directory; ignore EEXIST errors...
#if defined(WIN32) && ! defined (__CYGWIN__)
  if (mkdir(pathname))
#else
  if (mkdir(pathname, 0777))
#endif /* WIN32 */
    if (errno != EEXIST)
    {
      fltk3::alert("%s", strerror(errno));
      return;
    }

  // Show the new directory...
  directory(pathname);
}



/** Enable or disable the preview tile. 1 = enable preview, 0 = disable preview. */
void fltk3::FileChooser::preview(int e)
{
  previewButton->value(e);
  prefs_.set("preview", e);
  prefs_.flush();

  fltk3::Group *p = previewBox->parent();
  if (e) {
    int w = p->w() * 2 / 3;
    fileList->resize(fileList->x(), fileList->y(),
                     w, fileList->h());
    previewBox->resize(fileList->x()+w, previewBox->y(),
                       p->w()-w, previewBox->h());
    previewBox->show();
    update_preview();
  } else {
    fileList->resize(fileList->x(), fileList->y(),
                     p->w(), fileList->h());
    previewBox->resize(p->x()+p->w(), previewBox->y(),
                       0, previewBox->h());
    previewBox->hide();
  }
  p->init_sizes();

  fileList->parent()->redraw();
}


//
// 'fltk3::FileChooser::previewCB()' - Timeout handler for the preview box.
//

void
fltk3::FileChooser::previewCB(fltk3::FileChooser *fc) {	// I - File chooser
  fc->update_preview();
}


//
// 'fltk3::FileChooser::rescan()' - Rescan the current directory.
//

void
fltk3::FileChooser::rescan()
{
  char	pathname[FLTK3_PATH_MAX];		// New pathname for filename field


  // Clear the current filename
  strlcpy(pathname, directory_, sizeof(pathname));
  if (pathname[0] && pathname[strlen(pathname) - 1] != '/') {
    strlcat(pathname, "/", sizeof(pathname));
  }
//  puts("Setting fileName in rescan()");
  fileName->value(pathname);

  if (type_ & DIRECTORY)
    okButton->activate();
  else
    okButton->deactivate();

  // Build the file list...
  fileList->load(directory_, sort);
#ifndef WIN32	
  if (!showHiddenButton->value()) remove_hidden_files();
#endif
  // Update the preview box...
  update_preview();
}

//
/**  
  Rescan the current directory  without clearing the filename, 
  then select the file if it is in the list
*/
void fltk3::FileChooser::rescan_keep_filename()
{
  // if no filename was set, this is likely a diretory browser
  const char *fn = fileName->value();
  if (!fn || !*fn || fn[strlen(fn) - 1]=='/') {
    rescan();
    return;
  }

  int   i;
  char	pathname[FLTK3_PATH_MAX];		// New pathname for filename field
  strlcpy(pathname, fn, sizeof(pathname));

  // Build the file list...
  fileList->load(directory_, sort);
#ifndef WIN32	
  if (!showHiddenButton->value()) remove_hidden_files();
#endif
  // Update the preview box...
  update_preview();

  // and select the chosen file
  char found = 0;
  char *slash = strrchr(pathname, '/');
  if (slash) 
    slash++;
  else
    slash = pathname;
  for (i = 1; i <= fileList->size(); i ++)
#if defined(WIN32) || defined(__EMX__)
    if (strcasecmp(fileList->text(i), slash) == 0) {
#else
    if (strcmp(fileList->text(i), slash) == 0) {
#endif // WIN32 || __EMX__
      fileList->topline(i);
      fileList->select(i);
      found = 1;
      break;
    }

  // update OK button activity
  if (found || type_ & CREATE)
    okButton->activate();
  else
    okButton->deactivate();
}


//
// 'fltk3::FileChooser::showChoiceCB()' - Handle show selections.
//

void
fltk3::FileChooser::showChoiceCB()
{
  const char	*item,			// Selected item
		*patstart;		// Start of pattern
  char		*patend;		// End of pattern
  char		temp[FLTK3_PATH_MAX];		// Temporary string for pattern


  item = showChoice->text(showChoice->value());

  if (strcmp(item, custom_filter_label) == 0) {
    if ((item = fltk3::input("%s", pattern_, custom_filter_label)) != NULL) {
      strlcpy(pattern_, item, sizeof(pattern_));

      quote_pathname(temp, item, sizeof(temp));
      showChoice->add(temp);
      showChoice->value(showChoice->size() - 2);
    }
  } else if ((patstart = strchr(item, '(')) == NULL) {
    strlcpy(pattern_, item, sizeof(pattern_));
  } else {
    strlcpy(pattern_, patstart + 1, sizeof(pattern_));
    if ((patend = strrchr(pattern_, ')')) != NULL) *patend = '\0';
  }

  fileList->filter(pattern_);

  if (shown()) {
    // Rescan the directory...
    rescan_keep_filename();
  }
}


//
// 'fltk3::FileChooser::update_favorites()' - Update the favorites menu.
//

void
fltk3::FileChooser::update_favorites()
{
  int		i;			// Looping var
  char		pathname[FLTK3_PATH_MAX],		// Pathname
		menuname[2048];		// Menu name
  const char	*home;			// Home directory


  favoritesButton->clear();
  favoritesButton->add("bla");
  favoritesButton->clear();
  favoritesButton->add(add_favorites_label, fltk3::ALT + 'a', 0);
  favoritesButton->add(manage_favorites_label, fltk3::ALT + 'm', 0, 0, fltk3::MENU_DIVIDER);
  favoritesButton->add(filesystems_label, fltk3::ALT + 'f', 0);
    
  if ((home = getenv("HOME")) != NULL) {
    quote_pathname(menuname, home, sizeof(menuname));
    favoritesButton->add(menuname, fltk3::ALT + 'h', 0);
  }

  for (i = 0; i < 100; i ++) {
    sprintf(menuname, "favorite%02d", i);
    prefs_.get(menuname, pathname, "", sizeof(pathname));
    if (!pathname[0]) break;

    quote_pathname(menuname, pathname, sizeof(menuname));

    if (i < 10) favoritesButton->add(menuname, fltk3::ALT + '0' + i, 0);
    else favoritesButton->add(menuname);
  }

  if (i == 100) ((fltk3::MenuItem *)favoritesButton->menu())[0].deactivate();
}


//
// 'fltk3::FileChooser::update_preview()' - Update the preview box...
//

void
fltk3::FileChooser::update_preview()
{
  const char		*filename;	// Current filename
  const char            *newlabel = 0;  // New label text
  fltk3::SharedImage	*image = 0,     // New image
			*oldimage;	// Old image
  int			pbw, pbh;	// Width and height of preview box
  int			w, h;		// Width and height of preview image
  int                   set = 0;        // Set this flag as soon as a decent preview is found

  if (!previewButton->value()) return;

  filename = value();
  if (filename == NULL) {
    // no file name at all, so we have an empty preview
    set = 1;
  } else if (fltk3::filename_isdir(filename)) {
    // filename is a directory, show a folder icon
    newlabel = "@fileopen";
    set = 1;
  } else {
    struct stat s;
    if (fltk3::stat(filename, &s)==0) {
      if ((s.st_mode&S_IFMT)!=S_IFREG) {
        // this is no regular file, probably some kind of device
        newlabel = "@-3refresh"; // a cross
        set = 1;
      } else if (s.st_size==0) {
        // this file is emty
        newlabel = "<empty file>";
        set = 1;
      } else {
        // if this file is an image, try to load it
        window->cursor(fltk3::CURSOR_WAIT);
        fltk3::check();
        
        image = fltk3::SharedImage::get(filename);
        
        if (image) {
          window->cursor(fltk3::CURSOR_DEFAULT);
          fltk3::check();
          set = 1;
        }
      }
    }
  }

  oldimage = (fltk3::SharedImage *)previewBox->image();

  if (oldimage) oldimage->release();

  previewBox->image(0);

  if (!set) {
    FILE	*fp;
    int		bytes;
    char	*ptr;

    if (filename) fp = fltk3::fopen(filename, "rb");
    else fp = NULL;

    if (fp != NULL) {
      // Try reading the first 1k of data for a label...
      bytes = (int)fread(preview_text_, 1, sizeof(preview_text_) - 1, fp);
      preview_text_[bytes] = '\0';
      fclose(fp);
    } else {
      // Assume we can't read any data...
      preview_text_[0] = '\0';
    }

    window->cursor(fltk3::CURSOR_DEFAULT);
    fltk3::check();

    // Scan the buffer for printable UTF8 chars...
    for (ptr = preview_text_; *ptr; ptr++) {
      uchar c = uchar(*ptr);
      if ( (c&0x80)==0 ) {
        if (!isprint(c&255) && !isspace(c&255)) break;
      } else if ( (c&0xe0)==0xc0 ) {
        if (ptr[1] && (ptr[1]&0xc0)!=0x80) break;
        ptr++;
      } else if ( (c&0xf0)==0xe0 ) {
        if (ptr[1] && (ptr[1]&0xc0)!=0x80) break;
        ptr++;
        if (ptr[1] && (ptr[1]&0xc0)!=0x80) break;
        ptr++;
      } else if ( (c&0xf8)==0xf0 ) {
        if (ptr[1] && (ptr[1]&0xc0)!=0x80) break;
        ptr++;
        if (ptr[1] && (ptr[1]&0xc0)!=0x80) break;
        ptr++;
        if (ptr[1] && (ptr[1]&0xc0)!=0x80) break;
        ptr++;
      }
    } 
//         *ptr && (isprint(*ptr & 255) || isspace(*ptr & 255));
//	 ptr ++);

    // Scan the buffer for printable characters in 8 bit
    if (*ptr || ptr == preview_text_) {
      for (ptr = preview_text_;
         *ptr && (isprint(*ptr & 255) || isspace(*ptr & 255));
	 ptr ++);
    }

    if (*ptr || ptr == preview_text_) {
      // Non-printable file, just show a big ?...
      previewBox->label(filename ? "?" : 0);
      previewBox->align(fltk3::ALIGN_CLIP);
      previewBox->labelsize(75);
      previewBox->labelfont(fltk3::HELVETICA);
    } else {
      // Show the first 1k of text...
      int size = previewBox->h() / 20;
      if (size < 6) size = 6;
      else if (size > fltk3::NORMAL_SIZE) size = fltk3::NORMAL_SIZE;

      previewBox->label(preview_text_);
      previewBox->align((fltk3::Align)(fltk3::ALIGN_CLIP | fltk3::ALIGN_INSIDE |
                                   fltk3::ALIGN_LEFT | fltk3::ALIGN_TOP));
      previewBox->labelsize(size);
      previewBox->labelfont(fltk3::COURIER);
    }
  } else if (image) {
    pbw = previewBox->w() - 20;
    pbh = previewBox->h() - 20;

    if (image->w() > pbw || image->h() > pbh) {
      w   = pbw;
      h   = w * image->h() / image->w();

      if (h > pbh) {
	h = pbh;
	w = h * image->w() / image->h();
      }

      oldimage = (fltk3::SharedImage *)image->copy(w, h);
      previewBox->image((fltk3::Image *)oldimage);

      image->release();
    } else {
      previewBox->image((fltk3::Image *)image);
    }

    previewBox->align(fltk3::ALIGN_CLIP);
    previewBox->label(0);
  } else if (newlabel) {
    previewBox->label(newlabel);
    previewBox->align(fltk3::ALIGN_CLIP);
    previewBox->labelsize(newlabel[0]=='@'?75:12);
    previewBox->labelfont(fltk3::HELVETICA);
  }

  previewBox->redraw();
}


//
// 'fltk3::FileChooser::value()' - Return a selected filename.
//

const char *			// O - Filename or NULL
fltk3::FileChooser::value(int f)	// I - File number
{
  int		i;		// Looping var
  int		fcount;		// Number of selected files
  const char	*name;		// Current filename
  static char	pathname[FLTK3_PATH_MAX];	// Filename + directory


  name = fileName->value();

  if (!(type_ & MULTI)) {
    // Return the filename in the filename field...
    if (!name || !name[0]) return NULL;
    else return name;
  }

  // Return a filename from the list...
  for (i = 1, fcount = 0; i <= fileList->size(); i ++)
    if (fileList->selected(i)) {
      // See if this file is a selected file/directory...
      name = fileList->text(i);

      fcount ++;

      if (fcount == f) {
	if (directory_[0]) {
	  snprintf(pathname, sizeof(pathname), "%s/%s", directory_, name);
	} else {
	  strlcpy(pathname, name, sizeof(pathname));
	}

	return pathname;
      }
    }

  // If nothing is selected, use the filename field...
  if (!name || !name[0]) return NULL;
  else return name;
}


//
// 'fltk3::FileChooser::value()' - Set the current filename.
//

void
fltk3::FileChooser::value(const char *filename)
					// I - Filename + directory
{
  int	i,				// Looping var
  	fcount;				// Number of items in list
  char	*slash;				// Directory separator
  char	pathname[FLTK3_PATH_MAX];		// Local copy of filename


//  printf("fltk3::FileChooser::value(\"%s\")\n", filename == NULL ? "(null)" : filename);

  // See if the filename is the "My System" directory...
  if (filename == NULL || !filename[0]) {
    // Yes, just change the current directory...
    directory(filename);
    fileName->value("");
    okButton->deactivate();
    return;
  }

#ifdef WIN32
  // See if the filename contains backslashes...
  char	fixpath[FLTK3_PATH_MAX];			// Path with slashes converted
  if (strchr(filename, '\\')) {
    // Convert backslashes to slashes...
    strlcpy(fixpath, filename, sizeof(fixpath));

    for (slash = strchr(fixpath, '\\'); slash; slash = strchr(slash + 1, '\\'))
      *slash = '/';

    filename = fixpath;
  }
#endif // WIN32

  // See if there is a directory in there...
  fltk3::filename_absolute(pathname, sizeof(pathname), filename);

  if ((slash = strrchr(pathname, '/')) != NULL) {
    // Yes, change the display to the directory... 
    if (!fltk3::filename_isdir(pathname)) *slash++ = '\0';

    directory(pathname);
    if (*slash == '/') slash = pathname;
  } else {
    directory(".");
    slash = pathname;
  }

  // Set the input field to the absolute path...
  if (slash > pathname) slash[-1] = '/';

  fileName->value(pathname);
  fileName->position(0, (int)strlen(pathname));
  okButton->activate();

  // Then find the file in the file list and select it...
  fcount = fileList->size();

  fileList->deselect(0);
  fileList->redraw();

  for (i = 1; i <= fcount; i ++)
#if defined(WIN32) || defined(__EMX__)
    if (strcasecmp(fileList->text(i), slash) == 0) {
#else
    if (strcmp(fileList->text(i), slash) == 0) {
#endif // WIN32 || __EMX__
//      printf("Selecting line %d...\n", i);
      fileList->topline(i);
      fileList->select(i);
      break;
    }
}
  
void fltk3::FileChooser::show()
{
  window->hotspot(fileList);
  window->show();
  fltk3::flush();
  fltk3::cursor(fltk3::CURSOR_WAIT);
  rescan_keep_filename();
  fltk3::cursor(fltk3::CURSOR_DEFAULT);
  fileName->take_focus();
#ifdef WIN32
  showHiddenButton->hide();
#endif
}
  
void fltk3::FileChooser::showHidden(int value)
{
  if (value) {
    fileList->load(directory());
  } else {
    remove_hidden_files();
    fileList->redraw();
  }
}

void fltk3::FileChooser::remove_hidden_files()
{
  int count = fileList->size();
  for(int num = count; num >= 1; num--) {
    const char *p = fileList->text(num);
    if (*p == '.' && strcmp(p, "../") != 0) fileList->remove(num);
  }
  fileList->topline(1);
}
  


//
// 'compare_dirnames()' - Compare two directory names.
//

static int
compare_dirnames(const char *a, const char *b) {
  int alen, blen;

  // Get length of each string...
  alen = (int)strlen(a) - 1;
  blen = (int)strlen(b) - 1;

  if (alen < 0 || blen < 0) return alen - blen;

  // Check for trailing slashes...
  if (a[alen] != '/') alen ++;
  if (b[blen] != '/') blen ++;

  // If the lengths aren't the same, then return the difference...
  if (alen != blen) return alen - blen;

  // Do a comparison of the first N chars (alen == blen at this point)...
#ifdef WIN32
  return strncasecmp(a, b, alen);
#else
  return strncmp(a, b, alen);
#endif // WIN32
}


//
// 'quote_pathname()' - Quote a pathname for a menu.
//

static void
quote_pathname(char       *dst,		// O - Destination string
               const char *src,		// I - Source string
	       int        dstsize)	// I - Size of destination string
{
  dstsize --;

  while (*src && dstsize > 1) {
    if (*src == '\\') {
      // Convert backslash to forward slash...
      *dst++ = '\\';
      *dst++ = '/';
      src ++;
    } else {
      if (*src == '/') *dst++ = '\\';

      *dst++ = *src++;
    }
  }

  *dst = '\0';
}


//
// 'unquote_pathname()' - Unquote a pathname from a menu.
//

static void
unquote_pathname(char       *dst,	// O - Destination string
                 const char *src,	// I - Source string
	         int        dstsize)	// I - Size of destination string
{
  dstsize --;

  while (*src && dstsize > 1) {
    if (*src == '\\') src ++;
    *dst++ = *src++;
  }

  *dst = '\0';
}


//
// End of "$Id: FileChooser2.cxx 9552 2012-05-27 15:44:30Z manolo $".
//
